/* iony S-Link Protocol Device Driver for 2.4.x Linux Kernels
 * Copyright (c) 2002 Brian Behlendorf, behlendorf1@llnl.gov
 *
 * The S-Link driver is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * Full text for the GNU General Public License is available at
 * http://www.gnu.org/copyleft/gpl.html or by writing to the Free Software 
 * Foundation, Inc., 59 Temple Place Suite 330, Boston, MA  02111-1307,USA.
 */

#include "slink.h"


int slink_print_packet(char *buf, struct slink_packet *packet) {
  int i, deck, type, flag = 0;
  char *start = buf;

  TRACE("%p, %p", buf, packet);
  deck = (int)packet->data[0] & 0xFF;
  type = (int)packet->data[1] & 0xFF;

  if ((deck >= 0x90) && (deck < 0x96))
    buf += sprintf(buf, "Cmd %c ", ((deck - 0x90) % 3) + 0x41);
  else
    if ((deck >= 0x98) && (deck < 0x9E))
      buf += sprintf(buf, "Rpy %c ", ((deck - 0x98) % 3) + 0x41);
    else
      buf += sprintf(buf, "0x%-4X", deck);

  if ((deck >= 0x90) && (deck < 0x96)) {
    switch (type) {
      default:    PRINT_COMMAND(buf, "Unkn/Invalid");		break;
      case 0x00:  PRINT_COMMAND(buf, "Play");			break;
      case 0x01:  PRINT_COMMAND(buf, "Stop");			break;
      case 0x02:
      case 0x03:  PRINT_COMMAND(buf, "Pause");			break; 
      case 0x08:  PRINT_COMMAND(buf, "Next Track");		break; 
      case 0x09:  PRINT_COMMAND(buf, "Previous Track");		break;
      case 0x0E:  PRINT_COMMAND(buf, "Qry Strd Fields");	break;
      case 0x0F:  PRINT_COMMAND(buf, "Qry Deck Status");	break;
      case 0x10:  PRINT_COMMAND(buf, "Fast Forward");		break;
      case 0x11:  PRINT_COMMAND(buf, "Rewind");			break;
      case 0x12:  PRINT_COMMAND(buf, "F-Fast Forward");		break;
      case 0x13:  PRINT_COMMAND(buf, "Fast Rewind");		break;
      case 0x20:  PRINT_COMMAND(buf, "Remote Mode On");		break;
      case 0x21:  PRINT_COMMAND(buf, "Remote Mode Off");	break;
      case 0x22:  PRINT_COMMAND(buf, "Player Capacity");	break;
      case 0x25:  PRINT_COMMAND(buf, "Verbose Mode");		break;
      case 0x26:  PRINT_COMMAND(buf, "Brief Mode");		break;
      case 0x2E:  PRINT_COMMAND(buf, "Power On");		break;
      case 0x2F:  PRINT_COMMAND(buf, "Power Off");		break;
      case 0x40:  PRINT_COMMAND(buf, "Query Disc Memo");flag=3;	break;
      case 0x41:  PRINT_COMMAND(buf, "Qry Avail. Trks");	break;
      case 0x42:  PRINT_COMMAND(buf, "Qry Group Memo");		break;
      case 0x43:  PRINT_COMMAND(buf, "Qry Group Discs");	break;
      case 0x44:  PRINT_COMMAND(buf, "Query Disc");		break;
      case 0x45:  PRINT_COMMAND(buf, "Qry Disc/Trk");		break;
      case 0x48:  PRINT_COMMAND(buf, "Query Memo");		break;
      case 0x4A:  PRINT_COMMAND(buf, "Query BCD->Hex");		break;
      case 0x50:  PRINT_COMMAND(buf, "Play Disc/Trk");		break;
      case 0x51:  PRINT_COMMAND(buf, "Pause Disc/Trk");		break;
      case 0x5E:  PRINT_COMMAND(buf, "Fade Out");		break;
      case 0x5F:  PRINT_COMMAND(buf, "Fade In");		break;
      case 0x6A:  PRINT_COMMAND(buf, "Qry Deck Model");		break;
      case 0x72:  PRINT_COMMAND(buf, "Qry Deck Content");	break;
      case 0x80:  PRINT_COMMAND(buf, "Set Disc Memo");  flag=3;	break;
      case 0x81:  PRINT_COMMAND(buf, "Copy Disc Memo");		break;
      case 0x88:  PRINT_COMMAND(buf, "Set Disc Memo?");		break;
      case 0x90:  PRINT_COMMAND(buf, "Qry Disc Memo?");		break;
      case 0x91:  PRINT_COMMAND(buf, "Qry Disc Memo?");		break;
      case 0x92:  PRINT_COMMAND(buf, "Qry Disc Memo?");		break;
      /* The following are suspected to be valid but we lack details */
      case 0x1F: 
      case 0x23:
      case 0x24:
      case 0x27:
      case 0x29:
      case 0x30:
      case 0x31:
      case 0x46:
      case 0x47:
      case 0x63:
      case 0x64:
      case 0x74:  PRINT_COMMAND(buf, "Unkn/Valid");		break;
    }
  } else {
    if ((deck >= 0x98) && (deck < 0x9E)) {
      switch (type) {
        default:    PRINT_COMMAND(buf, "Unkn/Invalid");		break;
        case 0x00:  PRINT_COMMAND(buf, "Playing");		break;
        case 0x01:  PRINT_COMMAND(buf, "Stopped");		break;
        case 0x02:  PRINT_COMMAND(buf, "Paused");		break;
        case 0x05:  PRINT_COMMAND(buf, "No Disc");		break;
        case 0x06:  PRINT_COMMAND(buf, "Changing Disc");	break;
        case 0x08:  PRINT_COMMAND(buf, "Ready");		break;
        case 0x0C:  PRINT_COMMAND(buf, "29 Sec Remain");	break;
        case 0x0E:  PRINT_COMMAND(buf, "Duplicate Cmd");	break;
        case 0x0F:  PRINT_COMMAND(buf, "Invalid Cmd");		break;
        case 0x11:  PRINT_COMMAND(buf, "No Del. File?");	break;
        case 0x13:  PRINT_COMMAND(buf, "No Group?");		break; 
        case 0x14:  PRINT_COMMAND(buf, "D Not Loaded");		break;
        case 0x15:  PRINT_COMMAND(buf, "D/T Not Loaded");	break;
        case 0x18:  PRINT_COMMAND(buf, "Door Open");		break;  
        case 0x1D:  PRINT_COMMAND(buf, "Terminator");		break;
        case 0x1F:  PRINT_COMMAND(buf, "Disc Memo Set");flag=3;	break;
        case 0x2E:  PRINT_COMMAND(buf, "Power On");		break;
        case 0x2F:  PRINT_COMMAND(buf, "Power Off");		break;
        case 0x40:  PRINT_COMMAND(buf, "Memo");         flag=3;	break;
        case 0x41:  PRINT_COMMAND(buf, "Avail. Tracks");	break;
        case 0x42:  PRINT_COMMAND(buf, "Group");	flag=3;	break;
        case 0x44:
        case 0x45:  PRINT_COMMAND(buf, "Disc Memo Chk");flag=3;	break;
        case 0x47:  PRINT_COMMAND(buf, "CDText Disc,Fl?");	break;
        case 0x48:  PRINT_COMMAND(buf, "Memory");	flag=5;	break;
        case 0x50:  PRINT_COMMAND(buf, "Playing Disc");		break;
        case 0x51:  PRINT_COMMAND(buf, "Playing Disc at");	break;
        case 0x52:  PRINT_COMMAND(buf, "Displaying Disc");	break;
        case 0x53:  PRINT_COMMAND(buf, "Bad/Missing Dsc");	break;
        case 0x54:  PRINT_COMMAND(buf, "UnLoading Disc");	break;
        case 0x55:  PRINT_COMMAND(buf, "Empty");		break;
        case 0x58:  PRINT_COMMAND(buf, "Loading Disc");		break;
        case 0x5D:  PRINT_COMMAND(buf, "BCD->HEX");		break;
        case 0x60:  PRINT_COMMAND(buf, "Qry'd Disc");		break;
        case 0x61:  PRINT_COMMAND(buf, "Deck Size");		break;
        case 0x62:  PRINT_COMMAND(buf, "Qry'd Disc/Trk");	break;
        case 0x6A:  PRINT_COMMAND(buf, "Deck Model Id");flag=2;	break;
        case 0x70:  PRINT_COMMAND(buf, "Deck Mode");		break;
        case 0x71:  PRINT_COMMAND(buf, "Deck ???");		break;
        case 0x72:  PRINT_COMMAND(buf, "Deck Contents 1");	break;
        case 0x73:  PRINT_COMMAND(buf, "Deck Contents 2");	break;
        case 0x74:  PRINT_COMMAND(buf, "Deck Contents 3");	break;
        case 0x83:  PRINT_COMMAND(buf, "Door Closed");		break;
      }
    } else { buf += sprintf(buf, "0x%02X%12s|", type, ""); }
  }

  if (flag) {
    for (i = 0; i < flag; i++)
      buf += sprintf(buf,"%02X%s",(int)packet->data[i]&0xFF,(i==1)?"-":" ");

    for (i = flag; i < packet->size; i++)
      buf += sprintf(buf, " %c ", packet->data[i]);
  } else {
    for (i = 0; i < packet->size; i++)
      buf += sprintf(buf,"%02X%s",(int)packet->data[i]&0xFF,(i==1)?"-":" ");
  }

  return (int)(buf - start);
} /* slink_print_packet(t) */ 


int slink_read_proc(char *buf, char **start, off_t offset,
    int count, int *eof, void *data) {
  int i, index, len = 0, max = count - 80;
  struct list_head *ptr1, *ptr2;
  struct slink_packet *packet;
  struct slink_user *user;
  char swap[256];

  TRACE("%p, %p, %lu, %d, %p, %p", buf, start, offset, count, eof, data);
  for (i = 0; (i < SLINK_NO) && (len <= max); i++) {

    if (slink_table[i].dev == NULL)
      continue;

    down_interruptible(&(slink_table[i].user_sem));

    SPRINT(buf + len, "slink%d: major = %u io = 0x%lX irq = %d\n", i,
      slink_major, slink_table[i].dev->port->base,
      slink_table[i].dev->port->irq);
    SPRINT(buf + len, "interrupts = %ld\n", slink_table[i].interrupts);
    SPRINT(buf + len, "buffer     = %ld / %d\n",
      slink_table[i].irqs.size, IRQ_BUFFER_SIZE);
    SPRINT(buf + len, "flags      = 0x%02lX\nusers      = %lu\n\n",
      slink_table[i].flags, slink_table[i].users);
    SPRINT(buf + len, "packets:\n  read      = %ld\n  written   = %ld\n"
      "  corrected = %ld\n  corrupted = %ld\n  collision = %ld\n\n",
      slink_table[i].in_packets, slink_table[i].out_packets,
      slink_table[i].corrected_packets, 
      slink_table[i].corrupted_packets,
      slink_table[i].collision_packets);

    if (!list_empty(slink_table[i].user_queue.next)) {
      for (ptr1 = slink_table[i].user_queue.next;
           ptr1 != &(slink_table[i].user_queue); ptr1 = ptr1->next) {
        user = list_entry(ptr1, struct slink_user, list);
        SPRINT(buf + len, "user %d: flags = %02lX\n", user->id, user->flags);
        SPRINT(buf + len, "read queue: depth = %ld\n", user->in_queue_depth);
        index = 0; 

        if (list_empty(user->in_queue.next)) {
          SPRINT(buf + len, "  <read queue empty>\n");
        } else {
          for (ptr2 = user->in_queue.next;
               ptr2 != &(user->in_queue); ptr2 = ptr2->next) {
            packet = list_entry(ptr2, struct slink_packet, list);
            SPRINT(buf + len, "  %-2d ", index);
            slink_print_packet(swap, packet);
            SPRINT(buf + len, "%s", swap);
            SPRINT(buf + len, "\n");
            index++;
          }
        }

        SPRINT(buf + len, "\n");
      }    
    }

    SPRINT(buf+len,"write queue: depth = %ld\n",slink_table[i].out_queue_depth);
    index = 0;

    if (list_empty(slink_table[i].out_queue.next)) {
      SPRINT(buf + len, "  <write queue empty>\n");
    } else {
      for (ptr1 = slink_table[i].out_queue.next;
           ptr1 != &(slink_table[i].out_queue); ptr1 = ptr1->next) {
        packet = list_entry(ptr1, struct slink_packet, list);
        SPRINT(buf + len, "  %-2d ", index);
        slink_print_packet(swap, packet);
        SPRINT(buf + len, "%s", swap);
        SPRINT(buf + len, "\n");
        index++;
      }
    }

    SPRINT(buf + len, "\n");  
    up(&(slink_table[i].user_sem));
  }

  *eof = 1;
  return len;
} /* slink_read_proc() */


int slink_debug_read_proc(char *buf, char **start, off_t offset,
    int count, int *eof, void *data) {
  char *tail, *end;
  int tmp, len = 0;

  end = debug_buffer.buf + DEBUG_BUFFER_SIZE;
  if ((debug_buffer.size < DEBUG_BUFFER_SIZE) ||
      (debug_buffer.head == (end - 1)))
    tail = debug_buffer.buf;
  else
    tail = debug_buffer.head + 1;

  *start = buf;
  len = MIN(count, debug_buffer.size - offset);

/*
  printk(KERN_ERR "len:    %d\n", len);
  printk(KERN_ERR "offset: %ld\n", offset);
  printk(KERN_ERR "count:  %d\n", count);

  printk(KERN_ERR "buf:  %p\n", debug_buffer.buf);
  printk(KERN_ERR "head: %p\n", debug_buffer.head);
  printk(KERN_ERR "tail: %p\n", tail);
  printk(KERN_ERR "end:  %p\n", end);
  printk(KERN_ERR "size: %d\n", debug_buffer.size);
*/
  if ((tail + offset       <  end) && 
      (tail + offset + len >= end)) {
    /* This chunk will wrap handle it in two peices */
    tmp = end - (tail + offset);
    memcpy(buf, tail + offset, tmp);
    memcpy(buf + tmp, debug_buffer.buf, len - tmp);
  } else {
    /* This is a normal contigeous chunk just figure out if based on
     * the offset off the pointer if things wrap.
     */
    if (tail + offset < end) {
      memcpy(buf, tail + offset, len);
    } else {
      tmp = offset - (end - tail);
      memcpy(buf, debug_buffer.buf + tmp, len);
    }
  }

  *eof = 0;
  if (len < count)
    *eof = 1;

  return len;
} /* slink_debug_read_proc() */
