#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <slink.h>


int main(int argc, char *argv[]) {
  char rsp[1024], cmd[MAX_PACKET_SIZE];
  int i, n, seq = 0, slink, param, size = 2, flags = O_RDWR;
  unsigned long timeout = 0;
  struct timeval tval;
  fd_set rset;

  if (argc < 3) {
    fprintf(stdout, "Syntax: poke <device> <ctrl code> <parameters> "
	    "<timeout>\n        poke 0x93 0x50 0x04 0x02 30\n");
    return 1;
  }

  cmd[0] = (char)strtol(argv[1], NULL, 16);
  cmd[1] = (char)strtol(argv[2], NULL, 16);

  if (argc > 3)
    timeout = strtol(argv[argc - 1], NULL, 10);  

  if (timeout > 0)
    flags |= O_NONBLOCK;

  if ((slink = open("/dev/slink0", flags)) < 0) {
    fprintf(stdout, "Could not open device.\n");
    return 1;
  }

  for (param = 3; param < (argc - 1); param++) {
    cmd[param - 1] = (char)strtol(argv[param], NULL, 16);
    size++;
  }

  fprintf(stdout, "--------------- Sending ---------------\n[%d] - ", size);
  fflush(stdout);
  write(slink, cmd, size);

  for (i = 0; i < size; i++)
    fprintf(stdout, "0x%02X ", (int)cmd[i] & 0xFF); 

  fprintf(stdout, "\n");
  if (timeout == 0)
    return 0;

  fprintf(stdout, "\n-------------- Listening --------------\n");
  tval.tv_sec = timeout;
  tval.tv_usec = 0;

  /* Normally all data read from the driver will be the binary
   * data described in the command reference.  However, when
   * SLINK_VERBOSE is set the driver will first decode this binary
   * data in to a human readable form.
   */
  ioctl(slink, SLINK_VERBOSE, 1);

  /* Occasionally an extra unexpected interrupt will be generated
   * on the parralel port.  Normally the driver will fix this error
   * and return the corrected data.  However, if SLINK_ERROR_EXTRA_IRQ
   * is set the data will be discarded and an error will be returned
   * to the calling application.
   */
  ioctl(slink, SLINK_ERROR_EXTRA_IRQ, 0);

  while ((tval.tv_usec + tval.tv_sec * 1000000) > 0) {
    FD_ZERO(&rset);
    FD_SET(slink, &rset);

    if ((n = select(slink + 1, &rset, NULL, NULL, &tval)) < 0) {
      fprintf(stdout, "Error [%d] %s\n", errno, strerror(errno));
      return 1;
    }

    if (n == 0)
      continue;

    if (FD_ISSET(slink, &rset)) {
      if ((size = read(slink, rsp, 1024)) < 0) {
        if (errno != ENODATA)
          fprintf(stdout, "Error [%d] %s\n", errno, strerror(errno));
      } else {
        fprintf(stdout, "[%d] %s\n", seq, rsp);
        seq++;
      }
    }
  }

  close(slink);

  return 0;
} /* main() */
